"""Defines pipeline interface after it has been split into individual jobs based on a FileTree."""
from functools import lru_cache
from file_tree.file_tree import FileTree
from typing import Optional, Set, Dict, Collection, Any, Callable, Sequence, List, Tuple
from enum import Enum
from fsl.utils.fslsub import func_to_cmd
from loguru import logger
from pathlib import Path
from fsl_sub import submit, config
import re
import inspect
from contextlib import contextmanager
import fnmatch
from warnings import warn

class OutputMissing(IOError):
    """Raised if a job misses the output files."""
    pass


class InputMissingRun(IOError):
    """Raised if a job misses the input files when it is being launched."""
    pass
class InputMissingPipe(IOError):
    """Raised if a job misses input files while it is being added to the pipeline."""
    def __init__(self, target, missing, dependency_graph=None):
        self.target = target
        self.missing = missing
        if dependency_graph is None:
            dependency_graph = []
        self.dependency_graph = dependency_graph
        super().__init__("")

    def __str__(self, ):
        dependency = ' -> '.join([str(d) for d in self.dependency_graph])
        return f"{self.target} can not be added to pipeline as it is missing required input files: {self.missing} (dependency_graph: {dependency} )"


class RunMethod(Enum):
    """How to run the individual jobs."""

    local = 1
    submit = 2
    dask = 3

    def default():
        """Return default RunMethod, which is `submit` on a cluster and `local` otherwise."""
        return RunMethod.submit if config.has_queues() else RunMethod.local


class JobList:
    """Pipeline with a concrete set of jobs.
    
    Produced from a :class:`fsl_pipe.pipeline.Pipeline` based on the provided `FileTree`.

    The default `JobList` produced by `Pipeline.generate_jobs` is unsorted and 
    needs to be sorted by running `filter` before running.
    """

    def __init__(self, tree: FileTree, jobs: Sequence["SingleJob"], targets: Dict[str, "FileTarget"]):
        """Create a new concrete set of jobs."""
        self.tree = tree
        self.jobs = jobs
        self.targets = targets

    def __len__(self, ):
        return len(self.jobs)

    def filter(self, templates: Collection[str], overwrite=False, overwrite_dependencies=False, skip_missing=False) -> "JobList":
        """Filter out a subset of jobs.

        This does a total of three things:
        - Only include jobs needed to create the templates or file patterns given in `templates`.
        - Exclude jobs for which output files already exist, unless requested by `overwrite` or `overwrite_dependencies`.
        - Sort the jobs, so that dependencies are run before the jobs that depend on them.

        :param templates: target template keys; Any jobs required to produce those files will be kept in the returned pipeline
        :param overwrite: if True overwrite the files matching the template even if they already exist, defaults to False
        :param overwrite_dependencies: if True rerun any dependent jobs even if the output of those jobs already exists, defaults to False
        :return: Concrete pipeline with the jobs needed to produces `templates` (with optional overriding)
        """
        def add_target(target: "FileTarget", warn_msg=""):
            if target.exists() and (not overwrite or target.producer is None):
                return
            if target.producer is None:
                warn(f"No job found that produces {target.filename} {warn_msg}")
                return
            try:
                target.producer.add_to_jobs(all_jobs, overwrite=overwrite, overwrite_dependencies=overwrite_dependencies)
            except InputMissingPipe as e:
                if skip_missing:
                    return
                e.dependency_graph.insert(0, e.target)
                e.target = target.filename
                raise

        all_jobs: Tuple[List[SingleJob], Set[SingleJob]] = ([], set())
        if templates is None:
            for target in self.targets.values():
                add_target(target)
        else:
            for template in templates:
                if template in self.tree.template_keys():
                    for filename in self.tree.get_mult(template).data.flatten():
                        target = get_target(filename, self.targets, from_template=template)
                        add_target(target, f"(part of template {template})")
                else:
                    matches = get_matching_targets(template, self.targets)
                    if len(matches) == 0:
                        warn(f"No files were found that match {template}. It is not a template in the FileTree nor does it match any files that can be produced by this pipeline.")
                    for target in matches:
                        add_target(target, f"(matches {template})")
        return JobList(self.tree, all_jobs[0], self.targets)

    def batch(self, use_label=False, only_connected=False, split_on_ram=False, use_placeholders=True):
        """Batches groups of jobs into a single job.

        Two jobs will be batched if:

            1. There are no intermediate jobs that need to run between the two and cannot be included in the same batch (because of #3 below).
            2. One of the jobs depend on the other. This is disabled by default. Enable it by setting `only_connected` to True.
            3. The jobs are similar. Jobs are similar if they match all of the (enabled) properties listed below:

                a) They have the same batch label (set during pipeline creating using the `@pipe(batch=<batch label>)`. This can be enabled by setting the `use_label` keyword to True. If enabled, jobs without a batch label will never be merged.
                b) The jobs have the same submit parameters (e.g., "architecture", "coprocessor"), except for "jobtime" and "name". "jobram" can also be ignored by setting the `split_on_ram` keyword to True.
                c) The jobs have the same placeholder values. This will prevent jobs from different subjects to be merged with each other. It can be disabled by setting the `use_placeholders` to False. Alternatively, a subset of placeholder values could be considered by passing `use_placeholders` to a sequence of placeholders to consider (e.g., `use_placeholders=["subject", "session"]` will not merge jobs with different "subject" or "session" placeholder values, but will ignore any other placeholders).
        """
        all_flags = {key for job in self.jobs for key in job.submit_params.keys()}.difference(["jobtime", "name"])
        if not split_on_ram and "jobram" in all_flags:
            all_flags.remove("jobram")
        all_flags = sorted(all_flags)

        job_labels = []
        for job in self.jobs:
            label = []
            label.append(job.batch if use_label else 1)
            label.extend(job.submit_params.get(key, None) for key in all_flags)
            if use_placeholders:
                if use_placeholders == True:
                    parameters = job.set_parameters
                else:
                    parameters = {key: value for key, value in job.set_parameters.items() if key in use_placeholders}
                label.append(frozenset(parameters.items()))
            job_labels.append(tuple(label))

    
        to_batch = {
            tb: set(job for (job, label) in zip(self.jobs, job_labels) if label == tb)
            for tb in set(job_labels)
        }
        jobs = set(self.jobs)
        for to_batch, batching in to_batch.items():
            if to_batch[0] is None:
                # do not batch jobs without a label set (if `use_label` is True)
                continue
            other_jobs = jobs.difference(batching)
            new_jobs = batch_connected_jobs(batching, other_jobs)
            if not only_connected:
                new_jobs = batch_unconnected_jobs(new_jobs, other_jobs)
            jobs = set(new_jobs).union(other_jobs)
        return JobList(self.tree, jobs, self.targets)

    def report(self):
        """Produce tree reports with the relevant input/output templates."""
        from rich import print, color, tree
        if len(self.jobs) == 0:
            print("No jobs will be run.")
            return

        all_inputs = set.union(*[set(j.input_targets) for j in self.jobs])
        all_outputs = set.union(*[set(j.output_targets) for j in self.jobs])
        templates = {
            template
            for fns in (all_inputs, all_outputs)
            for target in fns
            for template in target.from_templates
        }

        def proc_line(tree_obj):
            for t in tree_obj.children:
                proc_line(t)
            line = tree_obj.label
            start = line.index('[cyan]') + 6
            end = line.index('[/cyan]')
            key = line[start:end]
            input_count = 0
            output_count = 0
            overwrite_count = 0
            exists_count = 0

            for fn in self.tree.get_mult(key).data.flatten():
                target = get_target(fn, self.targets)
                is_output = False
                is_input = False
                if target in all_outputs:
                    is_output = True
                if target in all_inputs:
                    is_input = True
                if target.exists() and is_output:
                    overwrite_count += 1
                elif is_output:
                    output_count += 1
                elif target.exists():
                    exists_count += 1
                    if is_input:
                        input_count += 1
            if (input_count + output_count + overwrite_count) == 0:
                return
            counter = "/".join([str(number) if color is None else f"[{color}]{number}[/{color}]" 
                for number, color in [
                    (overwrite_count, 'red'),
                    (output_count, 'yellow'),
                    (input_count, 'blue'),
                ]])
            tree_obj.label = f"{line} [{counter}]"

        for rich_obj in self.tree.filter_templates(templates).fill()._generate_rich_report():
            if isinstance(rich_obj, tree.Tree):
                proc_line(rich_obj)
            print(rich_obj)

    def run_datalad(self, ):
        """Make sure we can run the pipeline.

        Calls `datalad.get` on all input files and `datalad.unlock` on all output files.
        """
        input_targets = set()
        output_targets = set()

        for job_group in self.jobs.values():
            for j in job_group:
                input_targets.update(j.exists_before)
                output_targets.update(j.exists_after)
        
        input_targets.difference_update(output_targets)
        input_fns = [t.filename for t in input_targets]
        output_fns = [t.filename for t in output_targets if t.exists]

        from .datalad import get_dataset
        ds = get_dataset()
        ds.get(input_fns)
        ds.unlock(output_fns)

    def run(self, method: RunMethod=None, wait_for=()):
        """Run all the jobs that are required to produce the given templates.

        :param method: defines how to run the job
        :param wait_for: job IDs to wait for before running pipeline
        :return: set of all the jobs that need to be run to produce the template files
        """
        if method is None:
            method = RunMethod.default()
        elif not isinstance(method, RunMethod):
            method = RunMethod[method]
        if len(self.jobs) == 0:
            logger.info("No new jobs being run/submitted")
            return

        prev_count = 0
        run_jobs: Dict[SingleJob, Any] = {}
        while len(run_jobs) < len(self.jobs):
            for job in self.jobs:
                if job in run_jobs or any(j in self.jobs and j not in run_jobs for _, j in job.dependencies(only_missing=False)):
                    continue
                dependencies = [run_jobs[j] for _, j in job.dependencies(only_missing=False) if j in run_jobs]
                if len(dependencies) == 0:
                    dependencies = wait_for
                run_jobs[job] = job(
                    method=method,
                    wait_for=dependencies,
                )
            if len(run_jobs) == prev_count:
                raise ValueError("Unable to run/submit all jobs. Are there circular dependencies?")
            prev_count = len(run_jobs)


class JobParent:
    """Parent for `SingleJob` and `BatchJob`.
    """
    input_targets: Set["FileTarget"]
    output_targets: Set["FileTarget"]
    optional_targets: Set["FileTarget"]

    #@lru_cache(None)
    def dependencies(self, only_missing=True) -> Set[Optional["SingleJob"]]:
        """Return jobs on which this job depends.

        By default it only returns those related with missing input files.

        :param only_missing: set to False to also return dependencies that produce files that already exist on disk
        """
        jobs = set()
        for target in self.input_targets:
            if not (only_missing and target.exists()):
                jobs.add((target in self.optional_targets, target.producer))
        return jobs

    def missing_output(self, reset_cache=False):
        """
        Create a list of filenames that do not exist on disk.

        Optional outputs are not considered.

        :param reset_cache: set to True to not rely on an existing cached existence check
        """
        missing = set()
        for to_check in self.output_targets:
            if to_check in self.optional_targets:
                continue
            if reset_cache:
                to_check.reset_existence()
            if not to_check.exists():
                missing.add(to_check)
        return missing

    def missing_input(self, reset_cache=False, ignore_expected=False):
        """
        Create a list of filenames that do not exist on disk.

        Optional inputs are not considered.

        :param reset_cache: set to True to not rely on an existing cached existence check
        :param ignore_expected: set to True to ignore any missing files that have a job that will produce them in the pipeline
        """
        missing = set()
        for to_check in self.input_targets:
            if to_check in self.optional_targets:
                continue
            if reset_cache:
                to_check.reset_existence()
            if ignore_expected and to_check.producer is not None:
                continue
            if not to_check.exists():
                missing.add(to_check)
        return missing

    def add_to_jobs(self, all_jobs, overwrite=False, overwrite_dependencies=False):
        """Mark this job and all of its dependencies to run.

        This job is marked to run, if any of the output does not yet exist or overwrite is True.
        The dependencies are marked to run, if this job runs and either their output does not exist or overwrite_dependencies is True.

        :param all_jobs: list of individual jobs. This job and all required jobs are added to this list.
        :param overwrite: if True mark this job even if the output already exists, defaults to False
        :param overwrite_dependencies: if True mark the dependencies of this job even if their output already exists, defaults to False
        """
        if self in all_jobs[1]:
            return
        if not overwrite and len(self.missing_output()) == 0:
            return
        missing = self.missing_input(ignore_expected=True)
        if len(missing) > 0:
            raise InputMissingPipe(self, {m.filename for m in missing})
        subjobs = ([], set(all_jobs[1]))
        for optional, job in self.dependencies(only_missing=not overwrite_dependencies):
            try:
                if job is not None:
                    job.add_to_jobs(subjobs, overwrite_dependencies, overwrite_dependencies) and not optional
            except InputMissingPipe as e:
                if optional:
                    continue
                e.dependency_graph.insert(0, e.target)
                e.target = self
                raise
        all_jobs[0].extend(subjobs[0])
        all_jobs[1].update(subjobs[0])
        all_jobs[0].append(self)
        all_jobs[1].add(self)

    def __call__(self, method: RunMethod, wait_for=()):
        """Run the job."""
        if method == RunMethod.local:
            self.prepare_run()
            missing = self.missing_input()
            if len(missing) > 0:
                raise InputMissingRun(f"{self} can not run as it misses required input files: {missing}")
            logger.info(f"running {self}")
            self.job = self.function(**self.kwargs)
            missing = self.missing_output(reset_cache=True)
            if len(missing) > 0:
                raise OutputMissing(f"{self} failed to produce required output files: {missing}")
        elif method == RunMethod.submit:
            from .pipeline import Template
            self.prepare_run()
            local_submit = dict(self.submit_params)
            if 'logdir' not in local_submit:
                local_submit['logdir'] = 'log'
            Path(local_submit['logdir']).mkdir(exist_ok=True, parents=True)
            if 'name' not in local_submit:
                local_submit['name'] = self.job_name()
            cmd = func_to_cmd(self.function, (), self.kwargs, local_submit['logdir'], clean="on_success")
            if len(wait_for) == 0:
                wait_for = None
            self.job = submit(cmd, jobhold=wait_for, **local_submit)
            logger.debug(f"submitted {self} with job ID {self.job}")
        elif method == RunMethod.dask:
            import dask
            def dask_job(*other_jobs):
                if any(a != 0 for a in other_jobs):
                    logger.debug(f"{self} skipped because dependencies failed")
                    return 1
                try:
                    self(RunMethod.local)
                except Exception as e:
                    logger.exception(f"{self} failed: {e}")
                    return 1
                logger.debug(f"Running {self} using dask")
            self.job = dask.delayed(dask_job, name=str(self))(wait_for)
        return self.job

    def prepare_run(self):
        """
        Prepare to run this job.

        Steps:
        1. Creates output directory
        """
        for target in self.output_targets:
            target.filename.parent.mkdir(parents=True, exist_ok=True)

    def expected(self, ):
        """
        Return true if this job is expected to be able to run.
        """
        for target in self.input_targets:
            if not target.expected():
                return False
        return True


class SingleJob(JobParent):
    """A single job within a larger pipeline."""

    def __init__(self, function: Callable, kwargs, submit_params, input_targets, output_targets, optionals, set_parameters=None, batch=None):
        """
        Create a single job that can be run locally or submitted.

        :param function: python function
        :param kwargs: keyword arguments
        :param submit_params: instructions to submit job to cluster using `fsl_sub`
        :param input_targets: set of all filenames expected to exist before this job runs
        :param output_targets: set of all filenames expected to exist after this job runs
        :param optionals: set of filenames that are used to generate the dependency graph and yet might not exist
        :param set_parameters: dictionary with placeholder values used to distinguish this SingleJob with all those produces for the same function
        :param batch: label used to batch multiple jobs into one when submitting to the cluster
        """
        self.function = function
        if set_parameters is None:
            set_parameters = {}
        self.set_parameters = set_parameters
        self.kwargs = kwargs
        self.input_targets = input_targets
        self.output_targets = output_targets
        self.optional_targets = set(optionals)
        self.submit_params = submit_params
        self.batch = batch
        for target in self.input_targets:
            target.required_by.add(self)
        for target in self.output_targets:
            target.producer = self

    def job_name(self, ):
        """Return a string representation of this job."""
        if len(self.set_parameters) > 0:
            parameter_string = '_'.join([f"{key}-{value}" for key, value in self.set_parameters.items()])
            name = f"{self.function.__name__}_{parameter_string}"
        else:
            name = self.function.__name__
        value = re.sub(r'[^\w\s-]', '', name).strip().lower()
        return re.sub(r'[-\s]+', '-', value)

    def __repr__(self, ):
        """Print job as a function call."""
        parameter_string = ', '.join([f"{key}={value}" for key, value in self.set_parameters.items()])
        return f"{self.function.__name__}({parameter_string})"


def get_target(filename: Path, all_targets, from_template=None) -> "FileTarget":
    """
    Return a :class:`FileTarget` matching the input `filename`.

    If the `FileTarget` for `filename` is already in `all_targets`, it will be returned.
    Otherwise a new `FileTarget` will be added to `all_targets` and returned.

    :param filename: path to the input/intermediate/output filename
    :param all_targets: dictionary of all FileTargets
    :param from_template: template key used to obtain the filename
    """
    abs_path = Path(filename).absolute()
    if abs_path not in all_targets:
        all_targets[abs_path] = FileTarget(filename)
    if from_template is not None:
        all_targets[abs_path].from_templates.add(from_template)
    return all_targets[abs_path]


def get_matching_targets(pattern: str, all_targets) -> List["FileTarget"]:
    """
    Return all :class:`FileTarget` that match the given input pattern.

    :param pattern: filename definition supporting Unix shell-style wildcards as described in https://docs.python.org/3/library/fnmatch.html
    :param all_targets: dictionary of all FileTargets
    """
    abs_pattern = str(Path(pattern).absolute())
    matches = []
    for path, target in all_targets.items():
        if fnmatch.fnmatch(str(path), abs_pattern):
            matches.append(target)
    return matches


class FileTarget:
    """Input, intermediate, or output file.

    See :func:`get_target` for instructions on creating a new `FileTarget`.

    If a specific :class:`SingleJob` produces a filename, this can be indicated by setting :attr:`producer`:

    .. code-block:: python

        get_target(filename, all_targets).producer = job

    This will raise a `ValueError` if the filename is already produced by another job.

    If a specific :class:`SingleJob` requires a filename, this can be indicated by adding it to :attr:`required_by`:

    .. code-block:: python

        get_target(filename, all_targets).required_by.add(job)

    Filename existence can be checked using :meth:`exists`.
    This method uses caching. To reset the cache run :meth:`reset_existence`.

    To check if the filename can be created by this pipeline (or already exists) run :meth:`expected`.
    """

    def __init__(self, filename: Path):
        """
        Create a new target based on the provided filename.

        Do not call this method directly. 
        Instead use :func:`get_target`.

        :param filename: filename
        """
        self.filename = Path(filename)
        self._producer = None
        self.required_by: Set[SingleJob] = set()
        self.from_templates: Set[str] = set()

    def exists(self) -> bool:
        """
        Test whether file exists on disk.

        This function is lazy; once it has been checked once it will keep returning the same result.

        To reset use :meth:`reset_existence`.
        """
        if not hasattr(self, "_exists"):
            self._exists = self.filename.is_symlink() or self.filename.exists()
        return self._exists

    def reset_existence(self, ):
        """Ensure existence is checked again when running :meth:`exists`."""
        if hasattr(self, "_exists"):
            del self._exists

    def expected(self, ):
        """
        Return whether the file can be produced by the pipeline (or already exists).

        Returns False if the file does not exist and there is no way to produce it. Otherwise, True is returned
        """
        if self.exists():
            return True
        if self.producer is None:
            return False
        return self.producer.expected()

    @property
    def producer(self, ) -> SingleJob:
        """Job that can produce this file."""
        return self._producer

    @producer.setter
    def producer(self, new_value):
        if self._producer is not None:
            if self._producer is new_value:
                return
            raise ValueError(f"{self} can be produced by both {self.producer} and {new_value}")
        self._producer = new_value

    def __repr__(self, ):
        """Print filename of target."""
        return f"FileTarget({str(self.filename)})"


@contextmanager
def update_closure(*dictionaries, **kwargs):
    """Add the provided dictionaries to the globals dictionary.

    Inside the `with` block all the dictionary keys will be available in the local environment.
    After the `with` block ends the local environment will be cleaned up again.
    
    Use like this:

    .. code-block:: python

        def my_func()
            with add_to_globals({'a': 3}):
                print(a)  # prints 3
            print(a)  # raises a NameError (or prints whatever `a` is set to in the global environment)
    """
    func_globals = inspect.stack()[2].frame.f_globals

    # merge input dictionaries
    new_kwargs = {}
    for d in dictionaries:
        new_kwargs.update(d)
    new_kwargs.update(kwargs)

    to_restore = {}
    to_delete = set()
    for key in new_kwargs:
        if key in func_globals:
            to_restore[key] = func_globals[key]
        else:
            to_delete.add(key)

    func_globals.update(new_kwargs)
    yield  # return to run the code within the with-block

    # clean up the environment
    func_globals.update(to_restore)
    for key in to_delete:
        del func_globals[key]
    del func_globals


def call_batched_jobs(funcs, kwargs):
    for f, k in zip(funcs, kwargs):
        f(**k)

class BatchJob(JobParent):
    """Batched combination of multiple `SingleJob` instances that will be submitted together.
    """
    function = staticmethod(call_batched_jobs)

    def __init__(self, *sub_jobs):
        """Creates a new `BatchJob` from sub_jobs.

        `sub_jobs` can be either `SingleJob` or `BatchJob`. In either case they will be run in the order in which they are supplied.
        """
        self.torun: Sequence[SingleJob] = []
        for job in sub_jobs:
            if isinstance(job, BatchJob):
                self.torun.extend(job.torun)
            else:
                assert isinstance(job, SingleJob)
                self.torun.append(job)
        self.torun_set = set(self.torun)

        self.output_targets = {t for job in self.torun for t in job.output_targets}
        for t in self.output_targets:
            t._producer = self
        self.input_targets = {t for job in self.torun for t in job.input_targets if t not in self.output_targets}
        for t in self.input_targets:
            t.required_by.add(self)

        self.optional_targets = set.union(
            {
                t for t in self.output_targets
                if not any(t in job.output_targets and t not in job.optional_targets for job in self.torun)
            },
            {
                t for t in self.input_targets
                if not any(t in job.input_targets and t not in job.optional_targets for job in self.torun)
            },
        )

        as_tuples = set.intersection(*[{(key, value) for key, value in job.set_parameters.items()} for job in self.torun])
        self.set_parameters = {key: value for key, value in as_tuples}

    @property
    def batch(self, ):
        b = self.torun[0].batch
        if all(job.batch is not None and job.batch == b for job in self.torun):
            return b
        return None

    @property
    def kwargs(self, ):
        return {
            'funcs': [job.function for job in self.torun],
            'kwargs': [job.kwargs for job in self.torun],
        }

    @property
    def submit_params(self, ):
        def sum_value(name, *times):
            total = sum(int(t) for t in times if t is not None)
            return None if total == 0 else total

        def max_value(name, *rams):
            if any(r is not None for r in rams):
                return max(int(r) for r in rams if r is not None)
            return None

        def extend(name, *vars):
            all_vars = set()
            for v in vars:
                if v is not None:
                    all_vars.update(v)
            return list(all_vars)

        def unique_param(name, *flags):
            not_none = {f for f in flags if f is not None}
            if len(not_none) == 0:
                return None
            elif len(not_none) == 1:
                return not_none.pop()
            else:
                raise ValueError(f"Multiple different values for submit parameter flag {name} found. fsl-pipe does not know which one to choose out of: {not_none}")

        to_combine = {
            'jobtime': sum_value,
            'jobram': max_value,
            'threads': max_value,
            'export_vars': extend,
        }

        all_flags = {flag for job in self.torun for flag in job.submit_params.keys()}
        return {
            flag: to_combine.get(flag, unique_param)(flag, *[job.submit_params.get(flag, None) for job in self.torun])
            for flag in all_flags
        }

    def job_name(self, ):
        """Return a string representation of this job."""
        base = "" if self.batch is None else self.batch + '_'
        func_name = "-".join(sorted({job.function.__name__ for job in self.torun}))
        if len(self.set_parameters) > 0:
            parameter_string = '_'.join([f"{key}-{value}" for key, value in self.set_parameters.items()])
            name = f"{func_name}_{parameter_string}"
        else:
            name = func_name
        value = re.sub(r'[^\w\s-]', '', name).strip().lower()
        return base + re.sub(r'[-\s]+', '-', value)

    def __repr__(self, ):
        """Print job as a function call."""
        return f"Batch({self.torun})"


def has_dependencies(possible_parent: JobParent, possible_children: Set[JobParent], all_jobs):
    if possible_parent in possible_children:
        return True
    for _, child in possible_parent.dependencies(only_missing=False):
        if child in all_jobs and has_dependencies(child, possible_children, all_jobs):
            return True
    return False


def batch_connected_jobs(to_batch: Sequence[SingleJob], other_jobs: Collection[SingleJob]) -> List[JobParent]:
    """Iteratively combines two jobs 

    Two jobs will be batched if:
    1. They are both in `to_batch`
    2. One of the jobs depends on the other
    3. There is no other job that needs to run between these two jobs.
    Iteration through the jobs will continue until no further batching is possible.
    """
    other_jobs = set(other_jobs)
    batches = set(to_batch)
    nbatches_prev = len(batches) * 2
    while len(batches) != nbatches_prev:
        nbatches_prev = len(batches)
        for job in set(batches):
            if job not in batches:
                continue
            for _, dependency in job.dependencies(only_missing=False):
                if job in batches and dependency in batches:
                    # properties 1 and 2 are met; now checking property 3
                    all_jobs = set.union(batches, other_jobs)
                    indirect_dependency = any(has_dependencies(d, {dependency}, all_jobs) for _, d in job.dependencies(only_missing=False) if d in all_jobs and d != dependency)
                    if not indirect_dependency:
                        batches.remove(job)
                        batches.remove(dependency)
                        batches.add(BatchJob(dependency, job))
                        break
    return list(batches)


def batch_unconnected_jobs(to_batch: Sequence[JobParent], other_jobs: Collection[JobParent]) -> List[BatchJob]:
    """Batch jobs into generational sets

    All jobs that do not depend on any other job in `to_batch` will be added to the first generational batch.
    All jobs that only depend on the first generational batch will be added to the second.
    And so forth, until all jobs have been added to a generation.

    Only dependencies through `other_jobs` will be considered.
    """
    to_batch = set(to_batch)
    all_jobs = set.union(to_batch, other_jobs)

    generations = []

    while len(to_batch) > 0:
        generations.append(set())
        for job in to_batch:
            if not any(has_dependencies(dependency, to_batch, all_jobs) for _, dependency in job.dependencies(only_missing=False) if dependency in all_jobs):
                generations[-1].add(job)
        to_batch.difference_update(generations[-1])
    return [jobs.pop() if len(jobs) == 1 else BatchJob(*jobs) for jobs in generations]





"""
Declarative pipeline definition based on filetrees.

Typical usage:

.. code-block:: python

    from fsl_pipe import pipe, In, Out, Ref, Var

    @pipe
    def job(input_file: In, output_file: Out):
        # code to convert `input_file` to `output_file`

    pipe.cli()  # runs command line interface
"""
from .pipeline import Pipeline, pipe, In, Out, Ref, Var
from .job import update_closure
__version__ = '1.0.1'


"""
Defines the pipeline before a FileTree is provided.

A pipeline is a collection of functions with mapping from the function parameters to input/output/reference/placeholder basenames.
"""
from typing import List, Optional, Set, Tuple, Dict, Collection, Callable, Union, Any
import argparse
from dataclasses import dataclass
from file_tree import FileTree
from file_tree.file_tree import is_singular
import numpy as np
from copy import copy
import inspect
import xarray
from collections import namedtuple
from fsl_sub import submit
from file_tree.template import is_singular
from pathlib import Path
import graphviz
import fnmatch
try:
    import datalad
except ImportError:
    datalad = None

class Pipeline:
    """Collection of python functions forming a pipeline.

    You can either create a new pipeline (`from fsl_pipe import Pipeline; pipe = Pipeline()`) or use a pre-existing one (`from fsl_pipe import pipe`)

    Scripts are added to a pipeline by using the pipeline as a decorator (see :meth:`__call__`).

    To run the pipeline based on instructions from the command line run :meth:`pipe.cli(tree) <cli>`, 
    where tree is a FileTree defining the directory structure of the pipeline input & output files.

    :ivar scripts: list of :class:`PipedFunction`, which define the python functions forming the pipeline and their input/output templates
    """

    def __init__(self, scripts=None, default_output=None, default_submit=None):
        """Create a new empty pipeline."""
        if scripts is None:
            scripts = []
        if default_output is None:
            default_output = []
        if default_submit is None:
            default_submit = {}
        self.scripts: List[PipedFunction] = list(scripts)
        self.default_output: List[str] = list(default_output)
        check_submit_parameters(default_submit)
        self.default_submit = default_submit

    def __call__(self, function=None, *, kwargs=None, no_iter=None, placeholders=None, as_path=False, submit=None, batch=None):
        """Add a python function as a :class:`PipedFunction` to the pipeline.

        This is the main route through which jobs are added to the pipeline.

        .. code-block:: python

            from fsl_pipe import pipe, In, Out, Ref, Var

            @pipe(submit=dict(jobtime=40))
            def func(in_path: In, out_path: Out, ref_path: Ref, placeholder_key: Var):
                pass

        :param function: Optionally provide the function directly. This can be useful when not using `pipe` as a decorator, e.g.:
    
            .. code-block:: python

                from fsl_pipe import pipe, In, Out, Ref, Var
                from shutil import copyfile

                pipe(copyfile, kwargs={'src': In('template_in'), 'dst': Out('template_out')})

        :param kwargs: maps function keyword arguments to template names (In/Out/Ref), placeholder names (Var), or anything else. Templates or placeholders are replaced by their values based on the file-tree. Anything else is passed on unaltered.
        :param no_iter: optional set of parameters not to iterate over.
        :param as_path: if set to true, provide template filenames to the function as `pathlib.Path` objects instead of strings.
        :param placeholders: dictionary overriding the placeholders set in the filetree for this specific function. This can, for example, be used to run this function on a sub-set of subjects.
        :param submit: dictionary with the flags to pass on to `fsl_sub.submit`, when submitting jobs to a cluster queue.
        :param batch: label used to batch multiple jobs into one when submitting to the cluster
        """
        if submit is None:
            submit = {}
        submit_params = dict(**self.default_submit, **submit)
        check_submit_parameters(submit_params)
        def wrapper(func):
            self.scripts.append(PipedFunction(func, submit_params=submit_params, kwargs=kwargs, no_iter=no_iter, override=placeholders, as_path=as_path, batch=batch))
            return func
        if function is None:
            return wrapper
        wrapper(function)
        return function

    def generate_jobs(self, tree: FileTree):
        """
        Split the pipeline into individual jobs.

        Produces an unsorted `JobList`, which can be sorted by running `filter` on it.

        :param tree: set of templates for the input/output/reference files with all possible placeholder values
        """
        from .job import JobList, FileTarget
        all_targets: Dict[str, FileTarget] = {}
        jobs = []
        for script in self.scripts:
            jobs.extend(script.get_jobs(tree, all_targets))
        return JobList(tree, jobs, all_targets)

    def default_parser(self, tree: FileTree, parser: Optional[argparse.ArgumentParser]=None, include_vars=None, exclude_vars=()) -> argparse.ArgumentParser:
        """
        Add default `fsl-pipe` arguments to an argument parser (will create a new argument parser if needed).

        :param tree: `file_tree.FileTree` object describing the directory structure for the input/output files (defaults to datalad tree).
        :param parser: Optional argument parser that will be updated with the default `fsl-pipe` arguments (by default a new one is created).
        :param include_vars: if provided, only include expose placeholders in this list to the command line
        :param exclude_vars: exclude placeholders in this list from the command line
        :return: Argument parser with the default `fsl-pipe` arguments. If one was provided as input, that one will be returned.
        """
        from .job import RunMethod
        if parser is None:
            parser = argparse.ArgumentParser(description="Runs the pipeline")
        if len(self.scripts) == 0:
            raise ValueError("The pipeline does not contain any scripts...")
        templates = set.union(*[script.filter_templates(True, tree.template_keys()) for script in self.scripts])

        if len(self.default_output) == 0:
            default_output = list(sorted(templates))
        else:
            default_output = self.default_output

        parser.add_argument("templates", nargs="*", default=default_output,
                            help=f"Set to one or more template keys or file patterns (e.g., \"*.txt\"). Only these templates/files will be produced. If not provided all templates will be produced ({', '.join(default_output)}).")
        parser.add_argument("-m", '--pipeline_method', default=RunMethod.default().name,
                            choices=[m.name for m in RunMethod],
                            help=f"method used to run the jobs (default: {RunMethod.default().name})")
        parser.add_argument("-o", '--overwrite', action='store_true', help="If set overwrite any requested files.")
        parser.add_argument("-d", '--overwrite_dependencies', action='store_true',
                            help="If set also overwrites dependencies of requested files.")
        parser.add_argument("-j", '--job-hold', default='',
                            help='Place a hold on the whole pipeline until job has completed.')
        parser.add_argument('--skip-missing', default='',
                            help='If set skip running any jobs depending on missing data. This replaces the default behaviour of raising an error if any required input data is missing.')
        parser.add_argument("-q", '--quiet', action='store_true', help="Suppresses the report on what will be run (might speed up starting up the pipeline substantially).")
        parser.add_argument("-b", "--batch", action="store_true", help="Batch jobs based on submit parameters and pipeline labels before running them.")
        gui_help = "Start a GUI to select which output files should be produced by the pipeline."
        try:
            import fsl_pipe_gui
        except ModuleNotFoundError:
            gui_help = gui_help + " Missing requirement 'fsl-pipe-gui' to run the GUI. Please install this with `conda/pip install fsl-pipe-gui`."
        parser.add_argument("-g", "--gui", action="store_true", help=gui_help)
        if datalad is not None:
            parser.add_argument("--datalad", action='store_true',
                                help="run datalad get on all input data before running/submitting the jobs")
        for var, values in tree.placeholders.items():
            if not isinstance(var, str):
                continue
            if '/' in var:
                continue
            if var in exclude_vars:
                continue
            if include_vars is not None and var not in include_vars:
                continue
            if np.asarray(values).ndim == 1:
                default = ','.join([str(v) for v in values])
            else:
                default = str(values)
            parser.add_argument(f"--{var}", nargs='+', help=f"Use to set the possible values of {var} to the selected values (default: {default})")
        return parser

    def run_cli(self, args: argparse.Namespace, tree: FileTree):
        """
        Run the pipeline based on arguments extracted from the default argument parser.

        :param args: Namespace consisting of arguments extracted from the command line (produced by `argparse.ArgumentParser.parse_args`). This is expected to contain:

            - templates (defaults to `self.default_output`): list of which templates the pipeline should produce
            - pipeline_method: string with method used to run the jobs
            - overwrite: boolean, which if true overwrite any requested files
            - overwrite_dependencies: boolean, which if true also overwrites dependencies of requested files
            - job-hold: string with comma-separated list, which contains job(s) that will be waited for
            - skip-missing: whether to skip jobs depending on missing data instead of raising an error
            - datalad (defaults to False): if true, run datalad get on all input data before running/submitting the jobs
            - {placeholder} (defaults to values in FileTree): sequence of strings overwriting the possible values for a particular placeholder

        :param tree: Definition of pipeline input/output files
        """
        from .job import RunMethod
        for var in tree.placeholders:
            if isinstance(var, str) and getattr(args, var, None) is not None:
                tree.placeholders[var] = getattr(args, var)
        if args.gui:
            self.gui(tree, overwrite_dependencies=args.overwrite_dependencies, run_method=RunMethod[args.pipeline_method])
            return
        templates = set.union(*[script.filter_templates(output=True, all_templates=tree.template_keys()) for script in self.scripts])
        requested_templates = getattr(
            args, 'templates', 
            None if len(self.default_output) == 0 else self.default_output
        )
        not_in_tree = [template for template in requested_templates if template not in tree.template_keys()]
        if len(not_in_tree) > 0:
            raise ValueError(f"Requested templates {not_in_tree} are not defined in the FileTree")
        unrecognised_templates = set(requested_templates).difference(templates)
        if len(unrecognised_templates) > 0:
            raise ValueError(f"Requested templates {unrecognised_templates} cannot be produced by this pipeline")
        concrete = self.generate_jobs(tree)
        torun = concrete.filter(requested_templates, overwrite=args.overwrite, overwrite_dependencies=args.overwrite_dependencies, skip_missing=args.skip_missing)
        if not args.quiet:
            torun.report()
        if getattr(args, "batch", False):
            use_label = any(job.batch is not None for job in concrete.jobs)
            torun = torun.batch(use_label=use_label)
        if datalad is not None and args.datalad:
            torun.run_datalad()
        torun.run(RunMethod[args.pipeline_method], wait_for=() if args.job_hold == '' else args.job_hold.split(','))

    def cli(self, tree: Optional[FileTree]=None, include_vars=None, exclude_vars=(), cli_arguments=None):
        """
        Run the pipeline from the command line.

        :param tree: `file_tree.FileTree` object describing the directory structure for the input/output files (defaults to datalad tree).
        :param include_vars: if provided, only include expose variables in this list to the command line
        :param exclude_vars: exclude variables in this list from the command line
        """
        if tree is None:
            if datalad is None:
                raise ValueError("Argument 'tree' missing: please provide a FileTree describing the directory structure for the pipeline")
            from .datalad import get_tree
            try:
                tree = get_tree()
            except IOError:
                raise ValueError("Pipeline run outside of a datalad dataset, so a FileTree needs to be explicitly provided using the 'tree' argument")
            if tree is None:
                raise ValueError("No reference FileTree for pipeline found")

        parser = self.default_parser(tree=tree, include_vars=include_vars, exclude_vars=exclude_vars)
        args = parser.parse_args(cli_arguments)
        self.run_cli(args, tree)

    def gui(self, tree: FileTree, **kwargs):
        """
        Run the fsl-pipe-gui interface to select pipeline output.

        :param tree: `file_tree.FileTree` object describing the directory structure for the input/output files.
        :param overwrite_depencencies: set to True to overwrite dependencies
        :param run_method: overrides the default method to run the jobs
        """
        try:
            from fsl_pipe_gui import run_gui
        except ModuleNotFoundError:
            raise ModuleNotFoundError("'fsl-pipe-gui' needs to be installed to run the graphical user interface. Please run `pip/conda install fsl-pipe-gui` and try again.")
        run_gui(self, tree, **kwargs)

    def move_to_subtree(self, sub_tree=None, other_mappings=None):
        """
        Create a new pipeline that runs in a sub-tree of a larger tree rather than at the top level.

        :param sub_tree: name of the sub-tree in the FileTree
        :param other_mappings: other mappings between templates or placeholder names and their new values
        """
        if other_mappings is None:
            other_mappings = {}
        all_scripts = [script.move_to_subtree(sub_tree, other_mappings) for script in self.scripts]
        new_default_submit = _update_kwargs(self.default_submit, sub_tree, other_mappings)
        return Pipeline(all_scripts, [_update_key(key, sub_tree, other_mappings) for key in self.default_output], new_default_submit)

    @classmethod
    def merge(cls, pipelines: Collection["Pipeline"]):
        """
        Combine multiple pipelines into a single one.

        :param pipelines: pipelines containing part of the jobs
        :return: parent pipeline containing all of the jobs in pipelines
        """
        new_pipeline = Pipeline()
        for pipeline in pipelines:
            new_pipeline.scripts.extend([s.copy() for s in pipeline.scripts])
            new_pipeline.default_output.extend(pipeline.default_output)
        return new_pipeline

    def find(self, function: Union[str, Callable]):
        """
        Iterate over any pipeline scripts that run the provided function.

        Either the function itself or the name of the function can be given.
        """
        for script in list(self.scripts):
            if script.function == function or getattr(script.function, '__name__', None) == function:
                yield script

    def remove(self, function: Union[str, Callable]):
        """
        Remove any pipeline scripts that run the provided function from the pipeline.

        Either the function itself or the name of the function can be given.
        """
        for script in self.find(function):
            self.scripts.remove(script)

    def configure(self, kwargs):
        """
        Override the values passed on to the keyword arguments of all the scripts.

        Any keywords not expected by a script will be silently skipped for that script.
        """
        for script in self.scripts:
            try:
                script.configure(kwargs, allow_new_keywords=False, check=False)
            except KeyError:
                pass

    def add_to_graph(self, graph: graphviz.Digraph=None, tree: FileTree=None):
        """
        Add all the pipeline functions to the provided graph.

        :param graph: GraphViz graph object (will be altered)
        :param tree: concrete FileTree
        """
        if graph is None:
            graph = graphviz.Digraph("pipeline", format="svg")
        placeholder_color = {}
        for idx, script in enumerate(self.scripts):
            script.add_node(graph, idx, tree, placeholder_color)
        return graph


class PipedFunction:
    """Represents a function stored in a pipeline."""

    def __init__(self, function, submit_params: Dict, kwargs=None, no_iter=None, override=None, as_path=True, batch=None):
        """
        Wrap a function with additional information to run it in a pipeline.

        :param function: python function that will be run in pipeline.
        :param submit_params: parameters to submit job running python function to cluster using `fsl_sub`.
        :param kwargs: maps function keyword arguments to templates, variables, or actual values.
        :param templates: maps function keyword arguments to templates in the FileTree with additional information on whether the file is input/output/reference.
        :param no_iter: which parameters to not iterate over (i.e., they are passed to the function in an array).
        :param override: dictionary overriding the placeholders set in the filetree.
        :param as_path: whether to pass on pathlib.Path objects instead of strings to the functions (default: True).
        :param batch: label used to batch multiple jobs into one when submitting to the cluster.
        """
        self.function = function
        self.submit_params = submit_params
        if override is None:
            override = {}
        self.override = override
        self.as_path = as_path
        self.batch = batch

        self.all_kwargs: Dict[str, Any] = {}
        self.configure(function.__annotations__)

        if kwargs is not None:
            self.configure(kwargs)

        if no_iter is None:
            no_iter = set()
        elif isinstance(no_iter, str):
            no_iter = [no_iter]
        self.explicit_no_iter = set(no_iter)

    def copy(self, ):
        """Create a copy of this PipedFunction for pipeline merging."""
        new_script = copy(self)
        new_script.all_kwargs = dict(self.all_kwargs)
        self.explicit_no_iter = set(self.explicit_no_iter)
        return new_script

    @property
    def no_iter(self, ) -> Set[str]:
        """Sequence of placeholder names that should not be iterated over."""
        res = {key if value.key is None else value.key for key, value in self.placeholders.items() if value.no_iter}
        res.update(self.explicit_no_iter)
        return res

    def configure(self, kwargs, allow_new_keywords=True, check=True):
        """Override the values passed on to the keyword arguments of the script.

        :param kwargs: new placeholders/templates/values for keyword arguments
        :param allow_new_keywords: if set to False, don't allow new keywords
        """
        bad_keys = []
        signature = inspect.signature(self.function)
        has_kws = any(param.kind == param.VAR_KEYWORD for param in signature.parameters.values())

        if not allow_new_keywords:
            existing_kws = set(signature.parameters)
            if has_kws:
                existing_kws.update(self.all_kwargs.keys())
            if check:
                bad_keys = {key for key in kwargs.keys() if key not in signature.parameters}
                raise KeyError(f"Tried to configure {self} with keys that are not expected by the function: {bad_keys}")
        for key, value in kwargs.items():
            if not allow_new_keywords and key not in existing_kws:
                continue
            self.all_kwargs[key] = value

    @property
    def placeholders(self, ):
        """Return dictionary with placeholder values overriden for this function."""
        return {key: value for key, value in self.all_kwargs.items() if isinstance(value, PlaceHolder)}

    @property
    def templates(self, ):
        """Return dictionary with templates used as input, output, or reference for this function."""
        return {key: value for key, value in self.all_kwargs.items() if isinstance(value, Template)}

    @property
    def kwargs(self, ):
        """Return dictionary with keyword arguments that will be passed on to the function."""
        return {key: value for key, value in self.all_kwargs.items() if not any(
            isinstance(value, cls) for cls in (Template, PlaceHolder)
            )}

    def filter_templates(self, output=False, all_templates=None) -> Set[str]:
        """
        Find all input or output template keys.

        :param output: if set to True select the input rather than output templates
        :param all_templates: sequence of all possible templates (required if any Template keys use globbing)
        :return: set of input or output templates
        """
        res = set()
        for kwarg_key, template in self.templates.items():
            if ((template.input and not output) or
                (template.output and output)):
                template_key = kwarg_key if template.key is None else template.key

                # apply globbing if template key contains * or ?
                if '*' in template_key or '?' in template_key:
                    if all_templates is None:
                        raise ValueError(f"Template {template_key} uses globbing, but no template keys are not provided")
                    res.update(fnmatch.filter(all_templates, template_key))
                else:
                    res.add(template_key)
        return res

    def iter_over(self, tree: FileTree) -> Tuple[str, ...]:
        """
        Find all the placeholders that should be iterated over before calling the function.

        These are all the placeholders that affect the input templates, but are not part of `self.no_iter`.

        :param tree: set of templates with placeholder values
        :return: placeholder names to be iterated over sorted by name
        """
        tree = tree.update(**self.override)
        in_vars = self.all_placeholders(tree, False)
        in_vars_linked = {tree.placeholders.linkages.get(key, key) for key in in_vars}
        out_vars = {tree.placeholders.linkages.get(key, key) for key in self.all_placeholders(tree, True)}

        updated_no_iter = {
            tree.placeholders.linkages.get(
                tree.placeholders.find_key(key),
                tree.placeholders.find_key(key),
            ) for key in self.no_iter
        }
        all_in = in_vars_linked.union(updated_no_iter)
        if len(out_vars.difference(all_in)) > 0:
            raise ValueError(f"{self}: Output template depends on {out_vars.difference(all_in)}, which none of the input templates depend on")
        return tuple(sorted(
            {v for v in in_vars if tree.placeholders.linkages.get(v, v) not in updated_no_iter}, 
        ))
        
    def get_jobs(self, tree: FileTree, all_targets: Dict):
        """
        Get a list of all individual jobs defined by this function.

        :param tree: set of templates with placeholder values
        :param all_targets: mapping from filenames to Target objects used to match input/output filenames between jobs
        :return: sequence of jobs
        """
        from .job import SingleJob
        tree = tree.update(**self.override)
        to_iter = self.iter_over(tree)
        jobs = []
        done_kwargs = set()
        def freeze_value(value):
            if isinstance(value, xarray.DataArray):
                return str(value)
            elif isinstance(value, dict):
                return frozenset(value.items())
            else:
                return value

        for sub_tree in tree.iter_vars(to_iter):
            kwargs, in_fns, out_fns, optionals = _single_job_kwargs(self.all_kwargs, sub_tree, tree, all_targets, as_path=self.as_path)
            job_submit_params, _, _, _ = _single_job_kwargs(self.submit_params, sub_tree, tree, all_targets, as_path=False, is_submit_param=True)

            frozen_kwargs = frozenset([(key, freeze_value(value)) for (key, value) in kwargs.items()])
            if frozen_kwargs in done_kwargs:
                continue
            done_kwargs.add(frozen_kwargs)

            jobs.append(SingleJob(
                self.function,
                kwargs,
                job_submit_params,
                in_fns, out_fns, optionals,
                {name: sub_tree.placeholders[name] for name in to_iter},
                self.batch,
            ))
        return jobs
        
    def all_placeholders(self, tree: FileTree, output=False) -> Set[str]:
        """
        Identify the multi-valued placeholders affecting the input/output templates of this function.

        :param tree: set of templates with placeholder values
        :param output: if set to True returns the placeholders for the output than input templates
        :return: set of all placeholders that affect the input/output templates
        """
        res = set()
        for t in self.filter_templates(output, tree.template_keys()):
            res.update(tree.get_template(t).placeholders())
        if not output:
            for key, variable in self.placeholders.items():
                res.add(key if variable.key is None else variable.key)

        bad_keys = {key for key in res if key not in tree.placeholders}
        if len(bad_keys) > 0:
            raise ValueError(f"No value set for placeholders: {bad_keys}")

        _, only_multi = tree.placeholders.split()
        return {only_multi.find_key(key) for key in res if key in only_multi}


    def move_to_subtree(self, sub_tree=None, other_mappings=None):
        """
        Create a new wrapped function that runs in a sub-tree of a larger tree rather than at the top level.

        :param sub_tree: name of the sub-tree in the FileTree
        :param other_mappings: other mappings between templates or placeholder names and their new values
        """
        if other_mappings is None:
            other_mappings = {}
        new_script = copy(self)
        new_script.all_kwargs = _update_kwargs(self.all_kwargs, sub_tree, other_mappings)
        new_script.override = {_update_key(key, sub_tree, other_mappings): value for key, value in self.override.items()}
        new_script.submit_params = _update_kwargs(self.submit_params, sub_tree, other_mappings)
        return new_script

    def __repr__(self, ):
        """Print function name."""
        return f"PipedFunction({self.function.__name__})"

    def add_node(self, graph: graphviz.Graph, index, tree: FileTree, placeholder_color: Dict[str, str]):
        """
        Add a node representing this function for a pipeline diagram.

        :param graph: input pipeline diagram (will be altered)
        :param index: unique integer identifier to use within the graph
        """
        seaborn_colors = ['#0173b2', '#de8f05', '#029e73', '#d55e00', '#cc78bc', '#ca9161', '#fbafe4', '#949494', '#ece133', '#56b4e9']
        identifier = str(index) + self.function.__name__
        label = self.function.__name__
        if tree is not None:
            placs = self.iter_over(tree)
            if len(placs) > 0:
                for p in placs:
                    if p not in placeholder_color:
                        placeholder_color[p] = seaborn_colors[len(placeholder_color) % len(seaborn_colors)]
                plac_text = ';'.join(f'<font color="{placeholder_color[p]}">{p}</font>' for p in placs)
                label = fr"<{label}<br/>{plac_text}>"

        graph.node(identifier, label=label, color='red', shape='box')
        for output in (False, True):
            for t in self.filter_templates(output, tree.template_keys()):
                label = t
                if tree is not None:
                    all_plac = tree.get_template(t).required_placeholders() | tree.get_template(t).optional_placeholders()
                    multi_plac = sorted([p for p in all_plac if p not in tree.placeholders or not is_singular(tree.placeholders[p])])
                    for p in multi_plac:
                        if p not in placeholder_color:
                            placeholder_color[p] = seaborn_colors[len(placeholder_color) % len(seaborn_colors)]
                    multi_plac = ';'.join(f'<font color="{placeholder_color[p]}">{p}</font>' for p in multi_plac)
                    if len(multi_plac) > 0:
                        label = fr"<{t}<br/>{multi_plac}>"
                graph.node(t, label, color='blue')
                if output:
                    graph.edge(identifier, t)
                else:
                    graph.edge(t, identifier)

@dataclass
class Template(object):
    """Represents a keyword argument that will be mapped to a template path in the FileTree.
    
    :param key: template key in FileTree.
    :param input: Set to true if file should be considered as input (i.e., it should exist before the job is run).
    :param output: Set to true if file should be considered as output (i.e., it is expected to exist after the job is run).
    :param optional: Set to true if input/output file should be considered for creating the dependency graph, but still might not exist.
    """

    key: Optional[str] = None
    input: bool = False
    output: bool = False
    optional: bool = False

    def __call__(self, key=None, optional=False):
        """Override the template key and whether it is optional."""
        return Template(key, self.input, self.output, optional)

@dataclass
class PlaceHolder(object):
    """Represents a keyword argument that will be mapped to a placeholder value.
    
    :param key: placeholder key in FileTree.
    :param no_iter: if True the pipeline will not iterate over individual placeholder values, but rather run a single job with all possible placeholder values at once.
    """

    key: Optional[str] = None
    no_iter: bool = False

    def __call__(self, key=None, no_iter=None):
        """Override the Placeholder."""
        if no_iter is None:
            no_iter = self.no_iter
        return PlaceHolder(key, no_iter)


"""Use to mark keyword arguments that represent input filenames of the wrapped function.

These filenames are expected to exist before the function is run.

The actual filename is extracted from the FileTree based on the template key.
This template key will by default match the keyword argument name, but can be overriden by calling `In` (e.g., `In("other_key")`).
"""
In = Template(input=True)

"""Use to mark keyword arguments that represent output filenames of the wrapped function.

These filenames are expected to exist after the function is run.
An error is raised if they do not.

The actual filename is extracted from the FileTree based on the template key.
This template key will by default match the keyword argument name, but can be overriden by calling `In` (e.g., `In("other_key")`).
"""
Out = Template(output=True)

"""Use to mark keyword arguments that represent reference filenames of the wrapped function.

These filenames might or might not exist before or after the job is run.

The actual filename is extracted from the FileTree based on the template key.
This template key will by default match the keyword argument name, but can be overriden by calling `In` (e.g., `In("other_key")`).
"""
Ref = Template()

"""Use to mark keyword arguments that represent placeholder values in the pipeline.

Placeholder values are returned as :any:`PlaceholderValue` objects.
"""
Var = PlaceHolder()


def to_templates_dict(input_files=(), output_files=(), reference_files=()):
    """Convert a sequence of input/output/reference files into a template dictionary.

    Args:
        input_files (sequence, optional): Template keys representing input files. Defaults to ().
        output_files (sequence, optional): Template keys representing output files. Defaults to ().
        reference_files (sequence, optional): Template keys representing reference paths. Defaults to ().

    Raises:
        KeyError: If the same template key is used as more than one of the input/output/reference options

    Returns:
        dict: mapping of the keyword argument names to the Template objects
    """
    res = {}
    for files, cls in [
        (input_files, In),
        (output_files, Out),
        (reference_files, Ref),
    ]:
        for name in files:
            if isinstance(name, str):
                short_name = name.split('/')[-1]
            else:
                short_name, name = name
            if name in res:
                raise KeyError(f"Dual definition for template {name}")
            res[short_name] = cls(name)

    return res


"""
Named tuple representing a placeholder value with three fields.

:param key: string with placeholder key in pipeline.
:param index: integer with the index (or multiple integers with the indices if `no_iter` is set for this placeholder) of the placeholder values in list of all possible placeholder values.
:param value: actual placeholder value (or sequence of values if `no_iter is set for this placeholder)
"""
PlaceholderValue = namedtuple("PlaceholderValue", ["key", "index", "value"])


def _single_job_kwargs(wrapper_kwargs, single_tree: FileTree, full_tree: FileTree, all_targets, as_path=False, is_submit_param=False):
    """Create the keywords, input, and output filenames for a single job."""
    from .job import get_target
    final_dict = {}
    input_filenames = []
    output_filenames = []
    optional_filenames = []
    for kwarg_key, value in wrapper_kwargs.items():
        if isinstance(value, Template):
            key = kwarg_key if value.key is None else value.key

            use_glob = '*' in key or '?' in key
            iter_keys = fnmatch.filter(single_tree.template_keys(), key) if use_glob else [key]
            res = {}
            for proc_key in iter_keys:
                try:
                    proc_res = single_tree.get(proc_key)
                    is_xarray = False
                except KeyError:
                    proc_res = single_tree.get_mult(proc_key)
                    is_xarray = True
                if is_xarray:
                    if is_submit_param:
                        raise ValueError(f"Submit parameter {kwarg_key} is set to the template {key}, which has more than one possible value")
                    filenames = [get_target(fn, all_targets, proc_key) for fn in proc_res.data.flatten()]
                    if value.input:
                        input_filenames.extend(filenames)
                    if value.output:
                        output_filenames.extend(filenames)
                    if value.optional:
                        optional_filenames.extend(filenames)
                    proc_res = xarray.apply_ufunc(Path if as_path else str, proc_res, vectorize=True)
                else:
                    filename = get_target(proc_res, all_targets, proc_key)
                    if value.input:
                        input_filenames.append(filename)
                    if value.output:
                        output_filenames.append(filename)
                    if value.optional:
                        optional_filenames.append(filename)
                    proc_res = (Path if as_path else str)(proc_res)
                res[proc_key] = proc_res
            if not use_glob:
                res = res[key]
        elif isinstance(value, PlaceHolder):
            key = single_tree.placeholders.find_key(kwarg_key if value.key is None else value.key)
            plac_value = single_tree.placeholders[key]
            ref_values = list(full_tree.placeholders[key])

            if is_singular(plac_value):
                res = PlaceholderValue(key, ref_values.index(plac_value), plac_value)
                if is_submit_param:
                    res = res.value
            else:
                res = PlaceholderValue(key, tuple(ref_values.index(v) for v in plac_value), plac_value)
                if is_submit_param:
                    raise ValueError(f"Submit parameter {kwarg_key} is set to the placeholder {key}, which has more than one possible value")
        else:
            res = value
        final_dict[kwarg_key] = res
    return final_dict, input_filenames, output_filenames, optional_filenames


def check_submit_parameters(submit_params: Dict):
    """
    Check that the submit parameters are actually valid.
    
    Raises a ValueError if there are any submit parameters set not expected by fsl_sub.
    """
    signature = inspect.signature(submit)
    unrecognised = set(submit_params.keys()).difference(signature.parameters.keys())
    if len(unrecognised) > 0:
        raise ValueError(f"Unrecognised fsl_sub submit keywords: {unrecognised}")
    if 'jobhold' in submit_params:
        raise ValueError("Job-holds are managed by the fsl-pipe and cannot be set by the pipeline definition.")


def _update_key(key, sub_tree, other_mappings):
    """Update template key with sub_tree precursor, unless listed in other_mappings."""
    if key in other_mappings:
        return other_mappings[key]
    elif sub_tree is not None:
        return sub_tree + '/' + key
    else:
        return key


def _update_kwargs(input_dict, sub_tree, other_mappings):
    """Update placeholders and templates in input_dict with sub_tree precursor, unless listed in other_mappings."""
    kwargs = {}
    for key, value in input_dict.items():
        kwargs[key] = value
        if isinstance(value, PlaceHolder) or isinstance(value, Template):
            use_key = value.key if value.key is not None else key
            kwargs[key] = value(_update_key(use_key, sub_tree, other_mappings))
    return kwargs


pipe = Pipeline()


"""Define interface of fsl-pipe with datalad."""
from typing import IO
from file_tree import FileTree
import os.path as op
from functools import lru_cache


def get_tree(relative_dir='.', full_dir=None):
    """Get FileTree defined as "data.tree" at the top-level of datalad dataset."""
    import datalad.api as dl
    if full_dir is None:
        full_dir = relative_dir
    ds = get_dataset(full_dir)
    if ds is None:
        raise IOError(f"No dataset found in {full_dir}")
    if op.isfile(op.join(full_dir, 'data.tree')):
        tree = FileTree.read(op.join(full_dir, 'data.tree'), top_level=relative_dir)
    else:
        tree = FileTree.empty(top_level=relative_dir)
    for sub_ds in dl.subdatasets(dataset=ds):
        sub_path = sub_ds['gitmodule_name']
        tree_name = op.split(sub_path)[-1]
        tree.add_template(sub_path, "__" + tree_name)
        tree.add_subtree(get_tree(sub_path, sub_ds['path']), tree_name, "__" + tree_name)
    return tree


@lru_cache(None)
def get_dataset(directory='.'):
    """Get datalad dataset containing given directory (default: current directory)."""
    import datalad.api as dl
    ds = dl.Dataset(directory)
    if not ds.is_installed():
        return None
    return ds